using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.ComponentModel;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;
using System.IO;

namespace CSockets
{
    public class Client
    {
        private int port;
        private Socket ClientSocket;
        private bool isConnected;
        private IPAddress remoteIP;
        private List<Packet> ReceivedList;
        private AsyncCallback pfnCallBack;

        public event PropertyChangedEventHandler PropertyChange;
        public event PacketEventHandler PacketTransmit;

        public int Port
        {
            get { return port; }
            set
            {
                if (value > 65535 || value < 1)
                {
                    throw new Exception("Port must be between 1 and 65535.");
                }
                else if (IsConnected)
                {
                    throw new Exception("Can't change port while client is connected.");
                }
                else
                {
                    port = value;
                    if (PropertyChange != null)
                    {
                        PropertyChange(this, new PropertyChangedEventArgs("Port"));
                    }
                }
            }
        }
        public bool IsConnected
        {
            get { return isConnected; }
            private set
            {
                isConnected = value;
                if (PropertyChange != null)
                {
                    PropertyChange(this, new PropertyChangedEventArgs("IsConnected"));
                }
            }
        }
        public IPAddress RemoteIP
        {
            get { return remoteIP; }
            set
            {
                if (IsConnected)
                {
                    throw new Exception("Can't change remote IP Address while client is connected.");
                }
                else
                {
                    remoteIP = value;
                    if (PropertyChange != null)
                    {
                        PropertyChange(this, new PropertyChangedEventArgs("RemoteIP"));
                    }
                }
            }
        }

        public void Connect()
        {
            if (!IsConnected)
            {
                ClientSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                ClientSocket.BeginConnect(new IPEndPoint(remoteIP, Port), new AsyncCallback(OnClientConnect), null);
            }
        }

        public void Disconnect()
        {
            if (IsConnected)
            {
                ClientSocket.Close();
                IsConnected = false;
            }
        }

        private void OnClientConnect(IAsyncResult asyn)
        {
            try { ClientSocket.EndConnect(asyn); }
            catch { IsConnected = false; return; }
            IsConnected = true;
            WaitForData();
        }

        private void WaitForData()
        {
            Packet TempPacket = new Packet(ClientSocket, new byte[10240]);
            if (pfnCallBack == null)
            {
                pfnCallBack = new AsyncCallback(OnDataReceived);
            }
            try
            {
                ClientSocket.BeginReceive(TempPacket.Buffer, 0, TempPacket.Buffer.Length, SocketFlags.None, pfnCallBack, TempPacket);
            }
            catch (Exception)
            {
                IsConnected = false;
            }
        }

        private void OnDataReceived(IAsyncResult asyn)
        {
            int DataLength = 0;
            try
            {
                DataLength = ((Packet)asyn.AsyncState).StateSocket.EndReceive(asyn);
            }
            catch (SocketException)
            {
                IsConnected = false;
                return;
            }
            catch (ObjectDisposedException)
            {
                return;
            }
            byte[] Data = new byte[DataLength];
            for (int i = 0; i < DataLength; i++)
            {
                Data[i] = ((Packet)asyn.AsyncState).Buffer[i];
            }
            Packet TempPacket = new Packet(((Packet)asyn.AsyncState).StateSocket, Data);
            if (DataLength != 0)
            {
                ReceivedList.Add(TempPacket);
                if (PacketTransmit != null)
                {
                    PacketTransmit(this, new PacketEventArgs(TempPacket, PacketDirection.Incoming));
                }
            }
            WaitForData();
        }

        public void RemovePacketFromReceivedList(Packet p)
        {
            ReceivedList.Remove(p);
        }

        public void Send(byte[] buffer)
        {
            Packet p = new Packet(ClientSocket, buffer);
            ClientSocket.BeginSend(p.Buffer, 0, p.Buffer.Length, SocketFlags.None, new AsyncCallback(OnDataSent), p);
        }

        private void OnDataSent(IAsyncResult asyn)
        {
            ClientSocket.EndSend(asyn);
            if (PacketTransmit != null)
            {
                PacketTransmit(this, new PacketEventArgs((Packet)(asyn.AsyncState), PacketDirection.Outgoing));
            }
        }

        public Client(IPAddress RemoteIPAddress, int port)
        {
            IsConnected = false;
            RemoteIP = RemoteIPAddress;
            Port = port;
            ClientSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            ReceivedList = new List<Packet>();
        }
    }
}