using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.Timers;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters.Binary;

namespace CSockets
{
    public class Server
    {
        private int port;
        private Socket ServerSocket;
        private bool isRunning;
        private List<Socket> clientSockets;
        public List<Packet> Queue;
        private Timer FrTimer = new Timer(1000);
        private Timer SynTimer = new Timer(15000);
        private AsyncCallback pfnCallBack;

        private PacketHandler packetHandler;

        public event PropertyChangedEventHandler PropertyChange;
        public event ClientEventHandler ClientChange;
        public event PacketEventHandler PacketTransmit;

        public int Port
        {
            get { return port; }
            set
            {
                if (value > 65535 || value < 1)
                {
                    throw new Exception("Port must be between 1 and 65535.");
                }
                else if (IsRunning)
                {
                    throw new Exception("Can't change port while server is running.");
                }
                else
                {
                    port = value;
                    if (PropertyChange != null)
                    {
                        PropertyChange(this, new PropertyChangedEventArgs("Port"));
                    }
                }
            }
        }
        public bool IsRunning
        {
            get { return isRunning; }
            private set
            {
                isRunning = value;
                if (PropertyChange != null)
                {
                    PropertyChange(this, new PropertyChangedEventArgs("IsRunning"));
                }
            }
        }
        public List<Socket> ClientSockets
        {
            get { return clientSockets; }
            private set
            {
                clientSockets = value;
                if (PropertyChange != null)
                {
                    PropertyChange(this, new PropertyChangedEventArgs("ClientSockets"));
                }
            }
        }

        public void Start()
        {
            if (!IsRunning)
            {
                ServerSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
                ServerSocket.Bind(new IPEndPoint(IPAddress.Any, Port));
                ServerSocket.Listen(5);
                IsRunning = true;
                ServerSocket.BeginAccept(new AsyncCallback(OnClientConnect), null);
                FrTimer.Start();
                SynTimer.Start();
            }
        }

        public void Stop()
        {
            if (IsRunning)
            {
                foreach (Socket Sock in ClientSockets)
                {
                    ClientChange(this, new ClientEventArgs(Sock, ClientEvent.Disconnected));
                    Sock.Close();
                }
                ClientSockets.Clear();
                ServerSocket.Close();
                IsRunning = false;
            }
        }

        private void OnClientConnect(IAsyncResult asyn)
        {
            if (IsRunning)
            {
                Socket TempSocket = ServerSocket.EndAccept(asyn);
                ClientSockets.Add(TempSocket);
                if (ClientChange != null)
                {
                    ClientChange(this, new ClientEventArgs(TempSocket,ClientEvent.Connected));
                }
                WaitForData(TempSocket);
                ServerSocket.BeginAccept(new AsyncCallback(OnClientConnect), null);
            }
        }

        private void WaitForData(Socket ClientSocket)
        {
            Packet TempPacket = new Packet(ClientSocket, new byte[10240]);
            if (pfnCallBack == null)
            {
                pfnCallBack = new AsyncCallback(OnDataReceived);
            }
            try
            {
                ClientSocket.BeginReceive(TempPacket.Buffer, 0, TempPacket.Buffer.Length, SocketFlags.None, pfnCallBack, TempPacket);
            }
            catch(SocketException)
            {
                ClientSockets.Remove(ClientSocket);
                ClientChange(this, new ClientEventArgs(ClientSocket,ClientEvent.Disconnected));
            }
        }

        private void OnDataReceived(IAsyncResult asyn)
        {
            int DataLength = 0;
            try
            {
                DataLength = ((Packet)asyn.AsyncState).StateSocket.EndReceive(asyn);
            }
            catch (SocketException)
            {
                ClientSockets.Remove(((Packet)asyn.AsyncState).StateSocket);
                if (ClientChange != null)
                {
                    ClientChange(this, new ClientEventArgs(((Packet)asyn.AsyncState).StateSocket, ClientEvent.Disconnected));
                }
                return;
            }
            catch (ObjectDisposedException)
            {
                return;
            }
            byte[] Data = new byte[DataLength];
            for (int i = 0; i < DataLength; i++)
            {
                Data[i] = ((Packet)asyn.AsyncState).Buffer[i];
            }
            Packet TempPacket = new Packet(((Packet)asyn.AsyncState).StateSocket, Data);
            if (DataLength != 0)
            {
                Queue.Add(TempPacket);
                if (PacketTransmit != null)
                {
                    PacketTransmit(this, new PacketEventArgs((Packet)asyn.AsyncState,PacketDirection.Incoming));
                }
            }
            WaitForData(((Packet)asyn.AsyncState).StateSocket);
        }

        private void SendPackets(List<Packet> PacketList)
        {
            for (int i = 0; i < PacketList.Count; i++)
            {
                Send(PacketList[i]);
            }
        }

        public void Send(Packet PacketToSend)
        {
            PacketToSend.StateSocket.BeginSend(PacketToSend.Buffer, 0, PacketToSend.Buffer.Length, SocketFlags.None, new AsyncCallback(OnDataSent), PacketToSend);
        }

        private void OnDataSent(IAsyncResult asyn)
        {
            ((Packet)asyn.AsyncState).StateSocket.EndSend(asyn);
            if (PacketTransmit != null)
            {
                PacketTransmit(this, new PacketEventArgs((Packet)asyn.AsyncState, PacketDirection.Outgoing));
            }
        }

        void FTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            ProceedPacketFromQueue();
        }

        void SynTimer_Elapsed(object sender, ElapsedEventArgs e)
        {
            if (ClientSockets.Count > 0)
            {
                List<Packet> ListToSend = new List<Packet>();
                String StringToSend = "SYN";
                for (int i = 0; i < ClientSockets.Count; i++)
                {
                    ListToSend.Add(new Packet(ClientSockets[i], StringToSend));
                }
                SendPackets(ListToSend);
            }
        }

        private void ProceedPacketFromQueue()
        {
            if (Queue.Count > 0)
            {
                SendPackets(packetHandler(Queue[0]));
                Queue.RemoveAt(0);
            }
        }

        public Server(int port, PacketHandler Handler)
        {
            Port = port;
            ServerSocket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            IsRunning = false;
            packetHandler = Handler;
            ClientSockets = new List<Socket>();
            Queue = new List<Packet>();
            FrTimer.Elapsed += new ElapsedEventHandler(FTimer_Elapsed);
            SynTimer.Elapsed += new ElapsedEventHandler(SynTimer_Elapsed);
        }
    }
}